# noinspection PyUnresolvedReferences
import jpype.imports
# noinspection PyUnresolvedReferences
import org.gciatto.kt.math as _ktmath


BigInteger = _ktmath.BigInteger

BigDecimal = _ktmath.BigDecimal

MathContext = _ktmath.MathContext

RoundingMode = _ktmath.RoundingMode


def big_integer(value) -> BigInteger:
    return BigInteger.Companion.of(value)


def big_decimal(value) -> BigDecimal:
    return BigDecimal.Companion.of(value)



import unittest
from tuprolog.core import *
from tuprolog.core.visitors import AbstractTermVisitor


class TestVisitors(unittest.TestCase):

    def test_term_visitor(self):
        class MyVisitor(AbstractTermVisitor):
            def defaultValue(self, term):
                return 'a'
        visitor = MyVisitor()
        someTerm = Atom.of('b')
        for term in [Atom.of('a'), Integer.of(1), Var.of("X")]:
            self.assertEqual(term.accept(visitor), 'a')


if __name__ == '__main__':
    unittest.main()


from tuprolog import logger
# noinspection PyUnresolvedReferences
import jpype.imports
# noinspection PyUnresolvedReferences
import it.unibo.tuprolog.solve.channel as _channel
# noinspection PyUnresolvedReferences
import kotlin.jvm.functions as _kfunctions
from tuprolog.jvmutils import jmap, kfunction
from typing import Callable, Mapping
from functools import singledispatch

Channel = _channel.Channel

ChannelStore = _channel.ChannelStore

InputChannel = _channel.InputChannel

InputStore = _channel.InputStore

OutputChannel = _channel.OutputChannel

OutputStore = _channel.OutputStore

Listener = _kfunctions.Function1


def std_in() -> InputChannel:
    return InputChannel.stdIn()


@singledispatch
def input_channel(generator: Callable, availability_checker: Callable = None) -> InputChannel:
    if availability_checker is None:
        return InputChannel.of(kfunction(0)(generator))
    else:
        return InputChannel.of(kfunction(0)(generator), kfunction(0)(availability_checker))


@input_channel.register
def input_channel_from_string(string: str) -> InputChannel:
    return InputChannel.of(string)


def input_store(stdin: InputChannel = None, channels: Mapping[str, InputChannel] = None) -> InputStore:
    if stdin is None and channels is None:
        return InputStore.fromStandard()
    elif channels is None:
        return InputStore.fromStandard(stdin)
    else:
        cs = {InputStore.STDIN: stdin or std_in()}
        for k, v in channels:
            cs[k] = v
        return InputStore.of(jmap(cs))


def std_out() -> OutputChannel:
    return OutputChannel.stdOut()


def std_err() -> OutputChannel:
    return OutputChannel.stdErr()


def warn() -> OutputChannel:
    return OutputChannel.warn()


def output_channel(consumer: Callable) -> OutputChannel:
    return OutputChannel.of(kfunction(1)(consumer))


def output_store(
        stdout: OutputChannel = None,
        stderr: OutputChannel = None,
        warnings: OutputChannel = None,
        channels: Mapping[str, OutputChannel] = None
) -> OutputStore:
    if all((channel is None for channel in (stdout, stderr, warnings))):
        return OutputStore.fromStandard()
    elif channels is None:
        return OutputStore.fromStandard(
            stdout or std_out(),
            stderr or std_err(),
            warnings or warn()
        )
    else:
        cs = {
            OutputStore.STDOUT: stdout or std_out(),
            OutputStore.STDERR: stderr or std_err()
        }
        for k, v in channels:
            cs[k] = v
        return OutputStore.of(jmap(cs), warnings or warn())


def listener(consumer: Callable) -> Listener:
    return kfunction(1)(consumer)


logger.debug("Loaded JVM classes from it.unibo.tuprolog.solve.channel.*")



from tuprolog import logger
# noinspection PyUnresolvedReferences
import jpype.imports
# noinspection PyUnresolvedReferences
import it.unibo.tuprolog.solve.stdlib.rule as _rule
from tuprolog.solve import signature


KtAppend = _rule.Append

Arrow = _rule.Arrow

CurrentPrologFlag = _rule.CurrentPrologFlag

KtMember = _rule.Member

Not = _rule.Not

Once = _rule.Once

KtSemicolon = _rule.Semicolon

SetPrologFlag = _rule.SetPrologFlag


class Append:
    FUNCTOR = KtAppend.FUNCTOR

    ARITY = KtAppend.ARITY

    SIGNATURE = signature(FUNCTOR, ARITY)

    Base = KtAppend.Base.INSTANCE

    Recursive = KtAppend.Recursive.INSTANCE


Arrow = Arrow.INSTANCE

CurrentPrologFlag = CurrentPrologFlag.INSTANCE


class Member:
    FUNCTOR = KtMember.FUNCTOR

    ARITY = KtMember.ARITY

    SIGNATURE = signature(FUNCTOR, ARITY)

    Base = KtMember.Base.INSTANCE

    Recursive = KtMember.Recursive.INSTANCE


Not = Not.INSTANCE

Once = Once.INSTANCE


class Semicolon:
    FUNCTOR = KtSemicolon.FUNCTOR

    ARITY = KtSemicolon.ARITY

    SIGNATURE = signature(FUNCTOR, ARITY)

    class If:
        Then = KtSemicolon.If.Then.INSTANCE

        Else = KtSemicolon.If.Else.INSTANCE

    class Or:
        Left = KtSemicolon.Or.Left.INSTANCE

        Right = KtSemicolon.Or.Right.INSTANCE


logger.debug("Loaded JVM classes from it.unibo.tuprolog.solve.stdlib.rule.*")



from tuprolog import logger
# noinspection PyUnresolvedReferences
import jpype.imports
# noinspection PyUnresolvedReferences
import it.unibo.tuprolog.solve.stdlib.function as _function


AbsoluteValue = _function.AbsoluteValue.INSTANCE

Addition = _function.Addition.INSTANCE

ArcTangent = _function.ArcTangent.INSTANCE

BitwiseAnd = _function.BitwiseAnd.INSTANCE

BitwiseComplement = _function.BitwiseComplement.INSTANCE

BitwiseLeftShift = _function.BitwiseLeftShift.INSTANCE

BitwiseOr = _function.BitwiseOr.INSTANCE

BitwiseRightShift = _function.BitwiseRightShift.INSTANCE

Ceiling = _function.Ceiling.INSTANCE

Cosine = _function.Cosine.INSTANCE

Exponential = _function.Exponential.INSTANCE

Exponentiation = _function.Exponentiation.INSTANCE

FloatFractionalPart = _function.FloatFractionalPart.INSTANCE

FloatIntegerPart = _function.FloatIntegerPart.INSTANCE

FloatingPointDivision = _function.FloatingPointDivision.INSTANCE

Floor = _function.Floor.INSTANCE

IntegerDivision = _function.IntegerDivision.INSTANCE

Modulo = _function.Modulo.INSTANCE

Multiplication = _function.Multiplication.INSTANCE

NaturalLogarithm = _function.NaturalLogarithm.INSTANCE

Remainder = _function.Remainder.INSTANCE

Round = _function.Round.INSTANCE

Sign = _function.Sign.INSTANCE

SignReversal = _function.SignReversal.INSTANCE

Sine = _function.Sine.INSTANCE

SquareRoot = _function.SquareRoot.INSTANCE

Subtraction = _function.Subtraction.INSTANCE

ToFloat = _function.ToFloat.INSTANCE

Truncate = _function.Truncate.INSTANCE


logger.debug("Loaded JVM classes from it.unibo.tuprolog.solve.stdlib.function.*")



from tuprolog import logger
# noinspection PyUnresolvedReferences
import jpype.imports
# noinspection PyUnresolvedReferences
import it.unibo.tuprolog.core.parsing as _parsing
from tuprolog.core import Term, Struct, Constant, Var, Atom, Numeric, Integer, Real, Clause
from tuprolog.jvmutils import InputStream

TermParser = _parsing.TermParser

TermReader = _parsing.TermReader

ParseException = _parsing.ParseException

InvalidTermTypeException = _parsing.InvalidTermTypeException

# noinspection PyUnresolvedReferences
from tuprolog.core.operators import Operator, OperatorSet, DEFAULT_OPERATORS, EMPTY_OPERATORS

from typing import Union, Iterable


def term_parser(with_default_operators: bool=True, operators: OperatorSet=None) -> TermParser:
    if operators is None:
        if with_default_operators:
            return TermParser.getWithDefaultOperators()
        else:
            return TermParser.getWithNoOperator()
    else:
        if with_default_operators:
            return TermParser.withOperators(DEFAULT_OPERATORS.plus(operators))
        else:
            return TermParser.withOperators(operators)


def term_reader(with_default_operators: bool=True, operators: OperatorSet=None) -> TermParser:
    if operators is None:
        if with_default_operators:
            return TermReader.getWithDefaultOperators()
        else:
            return TermReader.getWithNoOperator()
    else:
        if with_default_operators:
            return TermReader.withOperators(DEFAULT_OPERATORS.plus(operators))
        else:
            return TermReader.withOperators(operators)


DEFAULT_TERM_PARSER = term_parser()

DEFAULT_TERM_READER = term_reader()


def parse_term(string: str, operators: OperatorSet=None) -> Term:
    if operators is None:
        return DEFAULT_TERM_PARSER.parseTerm(string)
    else:
        return DEFAULT_TERM_PARSER.parseTerm(string, operators)


def parse_struct(string: str, operators: OperatorSet=None) -> Struct:
    if operators is None:
        return DEFAULT_TERM_PARSER.parseStruct(string)
    else:
        return DEFAULT_TERM_PARSER.parseStruct(string, operators)


def parse_constant(string: str, operators: OperatorSet=None) -> Constant:
    if operators is None:
        return DEFAULT_TERM_PARSER.parseConstant(string)
    else:
        return DEFAULT_TERM_PARSER.parseConstant(string, operators)


def parse_var(string: str, operators: OperatorSet=None) -> Var:
    if operators is None:
        return DEFAULT_TERM_PARSER.parseVar(string)
    else:
        return DEFAULT_TERM_PARSER.parseVar(string, operators)


def parse_atom(string: str, operators: OperatorSet=None) -> Atom:
    if operators is None:
        return DEFAULT_TERM_PARSER.parseAtom(string)
    else:
        return DEFAULT_TERM_PARSER.parseAtom(string, operators)


def parse_numeric(string: str, operators: OperatorSet=None) -> Numeric:
    if operators is None:
        return DEFAULT_TERM_PARSER.parseNumeric(string)
    else:
        return DEFAULT_TERM_PARSER.parseNumeric(string, operators)


def parse_integer(string: str, operators: OperatorSet=None) -> Integer:
    if operators is None:
        return DEFAULT_TERM_PARSER.parseInteger(string)
    else:
        return DEFAULT_TERM_PARSER.parseInteger(string, operators)


def parse_real(string: str, operators: OperatorSet=None) -> Real:
    if operators is None:
        return DEFAULT_TERM_PARSER.parseReal(string)
    else:
        return DEFAULT_TERM_PARSER.parseReal(string, operators)


def parse_clause(string: str, operators: OperatorSet=None) -> Clause:
    if operators is None:
        return DEFAULT_TERM_PARSER.parseClause(string)
    else:
        return DEFAULT_TERM_PARSER.parseClause(string, operators)


def read_term(input: Union[InputStream, str], operators: OperatorSet=None) -> Term:
    input = ensure_input_steam(input)
    if operators is None:
        return DEFAULT_TERM_READER.readTerm(input)
    else:
        return DEFAULT_TERM_READER.readTerm(input, operators)


def read_terms(input: Union[InputStream, str], operators: OperatorSet=None) -> Iterable[Term]:
    input = ensure_input_steam(input)
    if operators is None:
        return DEFAULT_TERM_READER.readTerms(input)
    else:
        return DEFAULT_TERM_READER.readTerms(input, operators)


logger.debug("Loaded JVM classes from it.unibo.tuprolog.core.parsing.*")



from tuprolog import logger
# noinspection PyUnresolvedReferences
import jpype.imports
# noinspection PyUnresolvedReferences
import it.unibo.tuprolog.solve.flags as _flags
from tuprolog.core import Term
from tuprolog.jvmutils import kpair, jmap, jarray, Pair
from typing import Iterable, Union, Mapping


DoubleQuotes = _flags.DoubleQuotes

FlagStore = _flags.FlagStore

LastCallOptimization = _flags.LastCallOptimization

MaxArity = _flags.MaxArity

NotableFlag = _flags.NotableFlag

Unknown = _flags.Unknown

Flag = Pair

EMPTY_FLAG_STORE: FlagStore = FlagStore.EMPTY

DEFAULT_FLAG_STORE: FlagStore = FlagStore.DEFAULT

DoubleQuotes: NotableFlag = DoubleQuotes.INSTANCE

LastCallOptimization: NotableFlag = LastCallOptimization.INSTANCE

MaxArity: NotableFlag = MaxArity.INSTANCE

Unknown: NotableFlag = Unknown.INSTANCE


def flag(first: Union[str, NotableFlag, Iterable], value: Term = None) -> Flag:
    if isinstance(first, NotableFlag):
        if value is None:
            return first.toPair()
        else:
            return first.to(value)
    elif isinstance(first, str):
        if value is None:
            raise ValueError("Argument value is None")
        return Flag(first, value)
    elif isinstance(first, Iterable) and value is None:
        return kpair(first)
    else:
        raise ValueError("Argument first is not iterable nor str")


def flag_store(*flags: Union[NotableFlag, Flag, Iterable, FlagStore], **kwargs: Mapping[str, Term]):
    normal_flags = []
    notable_flags = []
    other_stores = []
    for f in flags:
        if isinstance(f, NotableFlag):
            notable_flags.append(f)
        elif isinstance(f, Flag):
            normal_flags.append(f)
        elif isinstance(f, FlagStore):
            other_stores.append(f)
        else:
            normal_flags.append(flag(f))
    store1 = FlagStore.of(jarray(NotableFlag)@notable_flags)
    store2 = FlagStore.of(jarray(Flag)@normal_flags)
    store3 = FlagStore.of(jmap(kwargs))
    store = store1.plus(store2).plus(store3)
    for s in other_stores:
        store = store.plus(s)
    return store


logger.debug("Loaded JVM classes from it.unibo.tuprolog.solve.flags.*")



from tuprolog import logger

import jpype

from tuprolog.core import TermComparator


@jpype.JImplements(TermComparator)
class AbstractTermComparator(object):
    @jpype.JOverride
    def equals(self, other):
        return self is other

    @jpype.JOverride
    def compare(self, first, second):
        raise NotImplementedError()


logger.debug("Loaded compatibility layer for JVM subtypes of " + str(TermComparator.class_.getName()))



# noinspection PyUnresolvedReferences
import jpype.imports
from tuprolog import logger
from tuprolog.core import Term
from tuprolog.solve import ExecutionContext
# noinspection PyUnresolvedReferences
import it.unibo.tuprolog.solve.exception.error as errors


MessageError = errors.MessageError


def message_error(content: Term, context: ExecutionContext, cause=None) -> MessageError:
    return MessageError.of(content, context, cause)


logger.debug("Loaded JVM classes from it.unibo.tuprolog.solve.exception.error.MessageError.*")



# noinspection PyUnresolvedReferences
import jpype.imports
from tuprolog import logger
from tuprolog.solve import ExecutionContext
# noinspection PyUnresolvedReferences
import it.unibo.tuprolog.solve.exception.error as errors
from tuprolog.solve.exception import LogicError


SystemError = errors.SyntaxError


def syntax_error(context: ExecutionContext, message: str, exception) -> SystemError:
    if isinstance(exception, LogicError):
        return SystemError.forUncaughtError(context, message, exception)
    return SystemError.forUncaughtException(context, message, exception)


logger.debug("Loaded JVM classes from it.unibo.tuprolog.solve.exception.error.SystemError.*")



from functools import reduce

from tuprolog import logger

# noinspection PyUnresolvedReferences
import jpype
# noinspection PyUnresolvedReferences
import jpype.imports

# noinspection PyUnresolvedReferences
from it.unibo.tuprolog.solve.sideffects import SideEffect
# noinspection PyUnresolvedReferences
from it.unibo.tuprolog.solve.sideffects import SideEffectFactory
# noinspection PyUnresolvedReferences
from it.unibo.tuprolog.solve.sideffects import SideEffectsBuilder

# noinspection PyUnresolvedReferences
from tuprolog.core import Clause, Term
# noinspection PyUnresolvedReferences
from tuprolog.core.operators import Operator, OperatorSet
# noinspection PyUnresolvedReferences
from tuprolog.solve.library import Library, AliasedLibrary, Libraries, libraries as new_libraries
# noinspection PyUnresolvedReferences
from tuprolog.solve.channel import InputChannel, OutputChannel
from tuprolog.pyutils import iterable_or_varargs, dict_or_keyword_args
from tuprolog.jvmutils import jlist, jmap

from typing import Mapping, Union, Iterable, Any


def _forward_iterable_or_varargs(callable, args, *callable_args):
    return iterable_or_varargs(args, lambda xs: callable(jlist(xs), *callable_args))


def _forward_dict_or_keywords(callable, dict, kwargs, *callable_args):
    return dict_or_keyword_args(dict, kwargs, lambda ds: callable(jmap(ds), *callable_args))


def reset_static_kb(*clauses: Union[Clause, Iterable[Clause]]) -> SideEffect.ResetStaticKb:
    return _forward_iterable_or_varargs(SideEffect.ResetStaticKb, clauses)


def add_static_clauses(*clauses: Union[Clause, Iterable[Clause]]) -> SideEffect.AddStaticClauses:
    return _forward_iterable_or_varargs(SideEffect.AddStaticClauses, clauses)


def remove_static_clauses(*clauses: Union[Clause, Iterable[Clause]]) -> SideEffect.RemoveStaticClauses:
    return _forward_iterable_or_varargs(SideEffect.RemoveStaticClauses, clauses)


def reset_dynamic_kb(*clauses: Union[Clause, Iterable[Clause]]) -> SideEffect.ResetDynamicKb:
    return _forward_iterable_or_varargs(SideEffect.ResetDynamicKb, clauses)


def add_dynamic_clauses(*clauses: Union[Clause, Iterable[Clause]]) -> SideEffect.AddDynamicClauses:
    return _forward_iterable_or_varargs(SideEffect.AddDynamicClauses, clauses)


def remove_dynamic_clauses(*clauses: Union[Clause, Iterable[Clause]]) -> SideEffect.RemoveDynamicClauses:
    return _forward_iterable_or_varargs(SideEffect.RemoveDynamicClauses, clauses)


def set_flags(flags: Mapping[str, Term] = {}, **kwargs: Term) -> SideEffect.SetFlags:
    return _forward_dict_or_keywords(SideEffect.SetFlags, flags, kwargs)


def reset_flags(flags: Mapping[str, Term] = {}, **kwargs: Term) -> SideEffect.ResetFlags:
    return _forward_dict_or_keywords(SideEffect.ResetFlags, flags, kwargs)


def clear_flags(*flag_names: str) -> SideEffect.ClearFlags:
    return _forward_iterable_or_varargs(SideEffect.ClearFlags, flag_names)


def load_library(alias: str, library: Library) -> SideEffect.LoadLibrary:
    return SideEffect.LoadLibrary(alias, library)


def unload_libraries(*aliases: str) -> SideEffect.UnloadLibraries:
    return _forward_iterable_or_varargs(SideEffect.UnloadLibraries, aliases)


def update_library(alias: str, library: Library) -> SideEffect.UpdateLibrary:
    return SideEffect.UpdateLibrary(alias, library)


def add_libraries(*libraries: Union[Libraries, AliasedLibrary]) -> SideEffect.AddLibraries:
    return SideEffect.AddLibraries(new_libraries(libraries))


def reset_libraries(*libraries: Union[Libraries, AliasedLibrary]) -> SideEffect.ResetLibraries:
    return SideEffect.ResetLibraries(new_libraries(libraries))


def set_operators(*operators: Union[Operator, Iterable[Operator]]) -> SideEffect.SetOperators:
    return _forward_iterable_or_varargs(SideEffect.SetOperators, operators)


def reset_operators(*operators: Union[Operator, Iterable[Operator]]) -> SideEffect.ResetOperators:
    return _forward_iterable_or_varargs(SideEffect.ResetOperators, operators)


def remove_operators(*operators: Union[Operator, Iterable[Operator]]) -> SideEffect.RemoveOperators:
    return _forward_iterable_or_varargs(SideEffect.RemoveOperators, operators)


def open_input_channels(
        channels: Mapping[str, InputChannel] = {},
        **kwargs: InputChannel
) -> SideEffect.OpenInputChannels:
    return _forward_dict_or_keywords(SideEffect.OpenInputChannels, channels, kwargs)


def reset_input_channels(
        channels: Mapping[str, InputChannel] = {},
        **kwargs: InputChannel
) -> SideEffect.ResetInputChannels:
    return _forward_dict_or_keywords(SideEffect.ResetInputChannels, channels, kwargs)


def close_input_channels(*names: Union[str, Iterable[str]]) -> SideEffect.CloseInputChannels:
    return _forward_iterable_or_varargs(SideEffect.CloseInputChannels, names)


def open_output_channels(
        channels: Mapping[str, OutputChannel] = {},
        **kwargs: OutputChannel
) -> SideEffect.OpenOutputChannels:
    return _forward_dict_or_keywords(SideEffect.OpenOutputChannels, channels, kwargs)


def reset_output_channels(
        channels: Mapping[str, OutputChannel] = {},
        **kwargs: OutputChannel
) -> SideEffect.ResetOutputChannels:
    return _forward_dict_or_keywords(SideEffect.ResetOutputChannels, channels, kwargs)


def close_output_channels(*names: Union[str, Iterable[str]]) -> SideEffect.CloseOutputChannels:
    return _forward_iterable_or_varargs(SideEffect.CloseOutputChannels, names)


def set_persistent_data(
        data: Mapping[str, Any] = {},
        **kwargs: Any
) -> SideEffect.SetPersistentData:
    return _forward_dict_or_keywords(SideEffect.SetPersistentData, data, kwargs, False)


def reset_persistent_data(
        data: Mapping[str, Any] = {},
        **kwargs: Any
) -> SideEffect.SetPersistentData:
    return _forward_dict_or_keywords(SideEffect.SetPersistentData, data, kwargs, True)


def set_durable_data(
        data: Mapping[str, Any] = {},
        **kwargs: Any
) -> SideEffect.SetDurableData:
    return _forward_dict_or_keywords(SideEffect.SetDurableData, data, kwargs, False)


def reset_durable_data(
        data: Mapping[str, Any] = {},
        **kwargs: Any
) -> SideEffect.SetDurableData:
    return _forward_dict_or_keywords(SideEffect.SetDurableData, data, kwargs, True)


def set_ephemeral_data(
        data: Mapping[str, Any] = {},
        **kwargs: Any
) -> SideEffect.SetEphemeralData:
    return _forward_dict_or_keywords(SideEffect.SetEphemeralData, data, kwargs, False)


def reset_ephemeral_data(
        data: Mapping[str, Any] = {},
        **kwargs: Any
) -> SideEffect.SetEphemeralData:
    return _forward_dict_or_keywords(SideEffect.SetEphemeralData, data, kwargs, True)


logger.debug("Loaded JVM classes from it.unibo.tuprolog.solve.sideffects.*")



from tuprolog import logger

# noinspection PyUnresolvedReferences
import jpype


@jpype.JImplementationFor("it.unibo.tuprolog.solve.SolverFactory")
class _KtSolverFactory:
    def __jclass_init__(self):
        pass


logger.debug("Configure Kotlin adapters for types in it.unibo.tuprolog.solve.*")



from tuprolog import Info
from tuprolog.core import *
from tuprolog.core.comparators import *
from tuprolog.core.exception import *
from tuprolog.core.formatters import *
from tuprolog.core.operators import *
from tuprolog.core.visitors import *
from tuprolog.core.parsing import *
from tuprolog.unify import *
from tuprolog.theory import *
from tuprolog.theory.parsing import *
from tuprolog.solve import *
from tuprolog.solve.exception import *
from tuprolog.solve.exception.error import *
from tuprolog.solve.exception.warning import *
from tuprolog.solve.channel import *
from tuprolog.solve.flags import *
from tuprolog.solve.function import *
from tuprolog.solve.primitive import *
from tuprolog.solve.sideffcts import *
from tuprolog.solve.library import *
from tuprolog.solve.library.exception import *
from tuprolog.solve.data import *
from tuprolog.solve.stdlib import *
from tuprolog.solve.stdlib.primitive import *
from tuprolog.solve.stdlib.rule import *
from tuprolog.solve.stdlib.function import *
from tuprolog.solve.classic import *
from tuprolog.solve.prolog import *

import code

code.interact(local=locals())

input()


from tuprolog import logger
# noinspection PyUnresolvedReferences
import jpype.imports
from tuprolog.core import Term, Clause, Integer
from tuprolog.solve import ExecutionContext, Signature, Solution, current_time_instant, MAX_TIMEOUT
# from tuprolog.solve.sideffcts import SideEffect
from tuprolog.pyutils import iterable_or_varargs
from tuprolog.jvmutils import jlist
from typing import List, Iterable, Callable
# noinspection PyUnresolvedReferences
import it.unibo.tuprolog.solve.primitive as _primitive


Primitive = _primitive.Primitive

Solve = _primitive.Solve

PrimitiveWrapper = _primitive.PrimitiveWrapper

SolveRequest = Solve.Request

SolveResponse = Solve.Request


@jpype.JImplements(Primitive)
class AbstractPrimitive(object):
    @jpype.JOverride
    def solve(self, request: SolveRequest) -> Iterable[SolveResponse]:
        raise NotImplementedError()


def primitive(callable: Callable[[SolveResponse], Iterable[SolveResponse]]) -> Primitive:
    class CallableToPrimitiveAdapter(AbstractPrimitive):
        def solve(self, request: SolveRequest) -> Iterable[SolveResponse]:
            return callable(request)

    return CallableToPrimitiveAdapter()


def solve_request(
        signature: Signature,
        arguments: List[Term],
        context: ExecutionContext,
        issuing_instant: int = current_time_instant(),
        max_duration: int = MAX_TIMEOUT
) -> SolveRequest:
    return SolveRequest(signature, arguments, context, issuing_instant, max_duration)


def solve_response(solution: Solution, *side_effects) -> SolveResponse:
    return iterable_or_varargs(side_effects, lambda ses: SolveResponse(solution, None, jlist(ses)))


def check_term_is_recursively_callable(request: SolveRequest, term: Term):
    return PrimitiveWrapper.checkTermIsRecursivelyCallable(request, term)


def ensuring_all_arguments_are_instantiated(request: SolveRequest) -> SolveRequest:
    return PrimitiveWrapper.ensuringAllArgumentsAreInstantiated(request)


def ensuring_procedure_has_permission(request: SolveRequest, signature: Signature, operation) -> SolveRequest:
    return PrimitiveWrapper.ensuringProcedureHasPermission(request, signature, operation)


def ensuring_clause_procedure_has_permission(request: SolveRequest, clause: Clause, operation) -> SolveRequest:
    return PrimitiveWrapper.ensuringClauseProcedureHasPermission(request, clause, operation)


def ensuring_argument_is_well_formed_indicator(request: SolveRequest, index: int) -> SolveRequest:
    return PrimitiveWrapper.ensuringArgumentIsWellFormedIndicator(request, index)


def not_implemented(request: SolveRequest, message: str) -> SolveResponse:
    return PrimitiveWrapper.notImplemented(request, message)


def not_supported(request: SolveRequest, message: str) -> SolveResponse:
    return PrimitiveWrapper.notSupported(request, message)


def ensuring_argument_is_well_formed_clause(request: SolveRequest, index: int) -> SolveRequest:
    return PrimitiveWrapper.ensuringArgumentIsWellFormedClause(request, index)


def ensuring_argument_is_instantiated(request: SolveRequest, index: int) -> SolveRequest:
    return PrimitiveWrapper.ensuringArgumentIsInstantiated(request, index)


def ensuring_argument_is_numeric(request: SolveRequest, index: int) -> SolveRequest:
    return PrimitiveWrapper.ensuringArgumentIsNumeric(request, index)


def ensuring_argument_is_struct(request: SolveRequest, index: int) -> SolveRequest:
    return PrimitiveWrapper.ensuringArgumentIsStruct(request, index)


def ensuring_argument_is_callable(request: SolveRequest, index: int) -> SolveRequest:
    return PrimitiveWrapper.ensuringArgumentIsCallable(request, index)


def ensuring_argument_is_variable(request: SolveRequest, index: int) -> SolveRequest:
    return PrimitiveWrapper.ensuringArgumentIsVariable(request, index)


def ensuring_argument_is_compound(request: SolveRequest, index: int) -> SolveRequest:
    return PrimitiveWrapper.ensuringArgumentIsCompound(request, index)


def ensuring_argument_is_atom(request: SolveRequest, index: int) -> SolveRequest:
    return PrimitiveWrapper.ensuringArgumentIsCompound(request, index)


def ensuring_argument_is_constant(request: SolveRequest, index: int) -> SolveRequest:
    return PrimitiveWrapper.ensuringArgumentIsConstant(request, index)


def ensuring_argument_is_ground(request: SolveRequest, index: int) -> SolveRequest:
    return PrimitiveWrapper.ensuringArgumentIsGround(request, index)


def ensuring_argument_is_char(request: SolveRequest, index: int) -> SolveRequest:
    return PrimitiveWrapper.ensuringArgumentIsChar(request, index)


def ensuring_argument_is_specifier(request: SolveRequest, index: int) -> SolveRequest:
    return PrimitiveWrapper.ensuringArgumentIsSpecifier(request, index)


def ensuring_argument_is_integer(request: SolveRequest, index: int) -> SolveRequest:
    return PrimitiveWrapper.ensuringArgumentIsInteger(request, index)


def ensuring_argument_is_list(request: SolveRequest, index: int) -> SolveRequest:
    return PrimitiveWrapper.ensuringArgumentIsList(request, index)


def ensuring_argument_is_arity(request: SolveRequest, index: int) -> SolveRequest:
    return PrimitiveWrapper.ensuringArgumentIsArity(request, index)


def ensuring_argument_is_non_negative_integer(request: SolveRequest, index: int) -> SolveRequest:
    return PrimitiveWrapper.ensuringArgumentIsNonNegativeInteger(request, index)


def is_character_code(integer: Integer) -> bool:
    return PrimitiveWrapper.isCharacterCode(integer)


def ensuring_term_is_char_code(request: SolveRequest, term: Term) -> SolveRequest:
    return PrimitiveWrapper.ensuringTermIsCharCode(request, term)


def ensuring_term_is_well_formed_list(request: SolveRequest, term: Term) -> SolveRequest:
    return PrimitiveWrapper.ensuringTermIsWellFormedList(request, term)


def ensuring_argument_is_char_code(request: SolveRequest, index: int) -> SolveRequest:
    return PrimitiveWrapper.ensuringArgumentIsCharCode(request, index)


logger.debug("Loaded JVM classes from it.unibo.tuprolog.solve.primitive.*")



from tuprolog import logger

# noinspection PyUnresolvedReferences
import jpype.imports

# noinspection PyUnresolvedReferences
import it.unibo.tuprolog.core.exception as _exceptions

TuPrologException = _exceptions.TuPrologException

SubstitutionException = _exceptions.TuPrologException

SubstitutionApplicationException = _exceptions.TuPrologException


logger.debug("Loaded JVM classes from it.unibo.tuprolog.core.exception.*")



from tuprolog import logger
# noinspection PyUnresolvedReferences
import jpype
# noinspection PyUnresolvedReferences
import jpype.imports
# noinspection PyUnresolvedReferences
import it.unibo.tuprolog.core.operators as _operators
from tuprolog.jvmutils import jiterable
from tuprolog.pyutils import iterable_or_varargs
from tuprolog.core import Atom, Integer, Struct, Term
from functools import singledispatch


Operator = _operators.Operator

OperatorSet = _operators.OperatorSet

Specifier = _operators.Specifier


@singledispatch
def operator(functor: str, specifier: Specifier, priority: int) -> Operator:
    return Operator(functor, specifier, priority)


@operator.register
def _(priority: Integer, specifier: Atom, functor: Atom) -> Operator:
    return Operator.fromTerms(priority, specifier, functor)


@operator.register
def _(term: Struct) -> Operator:
    return Operator.fromTerm(term)


def operator_set(*operators) -> OperatorSet:
    return iterable_or_varargs(operators, lambda os: OperatorSet(jiterable(os)))


@singledispatch
def specifier(name: str) -> Specifier:
    return Specifier.valueOf(name.upper())


@specifier.register
def _(term: Term) -> Specifier:
    return Specifier.fromTerm(term)


EMPTY_OPERATORS: OperatorSet = OperatorSet.EMPTY

DEFAULT_OPERATORS: OperatorSet = OperatorSet.DEFAULT

STANDARD_OPERATORS: OperatorSet = OperatorSet.STANDARD

XF: Specifier = Specifier.XF

YF: Specifier = Specifier.YF

FX: Specifier = Specifier.FX

FY: Specifier = Specifier.FY

XFX: Specifier = Specifier.XFX

XFY: Specifier = Specifier.XFY

YFX: Specifier = Specifier.YFX

OperatorSet.__add__ = lambda this, other: this.plus(other)
OperatorSet.__sub__ = lambda this, other: this.minus(other)


logger.debug("Loaded JVM classes from it.unibo.tuprolog.core.operators.*")



# noinspection PyUnresolvedReferences
import jpype.imports
from tuprolog import logger
from tuprolog.solve import ExecutionContext
# noinspection PyUnresolvedReferences
import it.unibo.tuprolog.solve.exception.error as errors


SyntaxError = errors.SyntaxError


def syntax_error(
        context: ExecutionContext,
        message: str
) -> SyntaxError:
    return SyntaxError.of(context, message)


def syntax_error_while_parsing_term(
        context: ExecutionContext,
        input: str,
        row: int,
        column: int,
        message: str
) -> SyntaxError:
    return SyntaxError.whileParsingTerm(context, input, row, column, message)


def syntax_error_while_parsing_clauses(
        context: ExecutionContext,
        input: str,
        index: int,
        row: int,
        column: int,
        message: str
) -> SyntaxError:
    return SyntaxError.whileParsingClauses(context, input, index, row, column, message)


def error_detector(text: str, line: int, column: int, message: str = None) -> str:
    return SyntaxError.errorDetector(text, line, column, message)


logger.debug("Loaded JVM classes from it.unibo.tuprolog.solve.exception.error.SyntaxError.*")



from tuprolog import logger
# noinspection PyUnresolvedReferences
import jpype.imports
# noinspection PyUnresolvedReferences
import it.unibo.tuprolog.solve.exception as _exceptions

HaltException = _exceptions.HaltException

LogicError = _exceptions.LogicError

ResolutionException = _exceptions.ResolutionException

TimeOutException = _exceptions.TimeOutException

Warning = _exceptions.Warning


logger.debug("Loaded JVM classes from it.unibo.tuprolog.solve.exception.*")



from tuprolog import logger
# noinspection PyUnresolvedReferences
import jpype.imports
# noinspection PyUnresolvedReferences
import it.unibo.tuprolog.theory as _theory
from typing import Iterable, Union
from tuprolog.core import Clause
from tuprolog.pyutils import iterable_or_varargs
from tuprolog.jvmutils import jiterable

Theory = _theory.Theory

MutableTheory = _theory.MutableTheory

RetractResult = _theory.RetractResult


def theory(*clauses: Union[Clause, Iterable[Clause]]) -> Theory:
    return iterable_or_varargs(clauses, lambda ts: Theory.of(jiterable(ts)))


def mutable_theory(*clauses: Union[Clause, Iterable[Clause]]) -> MutableTheory:
    return iterable_or_varargs(clauses, lambda ts: MutableTheory.of(jiterable(ts)))


logger.debug("Loaded JVM classes from it.unibo.tuprolog.theory.*")



from tuprolog import logger

import jpype


@jpype.JImplementationFor("it.unibo.tuprolog.core.Term")
class _KtTerm:
    def __jclass_init__(self):
        pass

    def __getitem__(self, item, *items):
        return self.get(item, *items)

    @property
    def variables(self):
        return self.getVariables()


@jpype.JImplementationFor("it.unibo.tuprolog.core.operators.Operator")
class _KtOperator:
    def __jclass_init__(self):
        pass

    @property
    def functor(self):
        return self.getFunctor()

    @property
    def specifier(self):
        return self.getSpecifier()

    @property
    def priority(self):
        return self.getPriority()


@jpype.JImplementationFor("it.unibo.tuprolog.core.Struct")
class _KtStruct:
    def __jclass_init__(self):
        pass

    @property
    def functor(self):
        return self.getFunctor()

    @property
    def args(self):
        return self.getArgs()

    @property
    def arity(self):
        return self.getArity()


@jpype.JImplementationFor("it.unibo.tuprolog.core.Clause")
class _KtClause:
    def __jclass_init__(self):
        pass

    @property
    def head(self):
        return self.getHead()

    @property
    def body(self):
        return self.getBody()

    @property
    def is_well_formed(self):
        return self.isWellFormed()


logger.debug("Configure Kotlin adapters for types in it.unibo.tuprolog.core.*")
